package com.laguerre.kan;

import java.util.ArrayList;
import com.laguerre.kan.SceneConductor.ISurfaceChangedListener;
import android.opengl.GLES20;
import android.opengl.Matrix;

public final class Camera extends Component implements ISurfaceChangedListener {
	public static class WrongFrameBufferHandleException extends RuntimeException {
		private static final long serialVersionUID = 9066562408218138881L;

		public WrongFrameBufferHandleException(String msg) {
			super(msg);
		}
	}

	public static class FrameBufferIncompleteException extends RuntimeException {
		private static final long serialVersionUID = -8348255784559121128L;

		public FrameBufferIncompleteException(String msg) {
			super(msg);
		}
	}

	private static Camera currentCamera = null;
	public static Camera getCurrent() {
		return currentCamera;
	}

	private boolean enabled;
	private RenderTexture renderTexture;
	private int fbHandle;
	private int depthFbHandle;
	private boolean glParamsInitialized = false;
	private int passNum;
	private boolean needsPass;
	private final ArrayList<SceneObject> multiPassSOb = new ArrayList<SceneObject>();

	private final float[] backgroundColor = new float[4];
	private float fovY = 45f;
	private float nearClipPlane = 0.01f;
	private float farClipPlane = 20f;

	private final int[] viewSize = new int[2];
	private final float[] viewMatrix = new float[16];
	private final float[] projMatrix = new float[16];
	private final float[] viewProjMatrix = new float[16];

	private final CommonRenderingPrms renderingPrms = new CommonRenderingPrms();

	Camera() {
		this.enabled = true;
		this.renderTexture = null;
		this.fbHandle = -1;
		this.depthFbHandle = -1;

		setBackgroundColor(0.3f, 0.3f, 0.3f, 1.0f);
		SurfaceInfo.getInstance().getSurfaceSize(this.viewSize);
	}

	public void setAsCurrent() {
		if (currentCamera != null) {
			SceneConductor.getInstance().removeSurfaceChangedListener(this);
		}
		currentCamera = this;
		SceneConductor.getInstance().addSurfaceChangedListener(this);
	}

	public void setEnabled(boolean enabling) {
		this.enabled = enabling;
	}

	public boolean isEnabled() {
		return this.enabled;
	}

	public void setRenderTexture(RenderTexture renderTexture) {
		if (this.renderTexture != renderTexture) {
			this.renderTexture = renderTexture;
			setupFBO();
		}
	}

	private void setupFBO() {
		final int[] names = { -1 };

		// FB
		GLES20.glGenFramebuffers(1, names, 0);
		this.fbHandle = names[0];

		// depth
		GLES20.glGenRenderbuffers(1, names, 0);
		this.depthFbHandle = names[0];

		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, this.depthFbHandle);
		GLES20.glRenderbufferStorage(
				GLES20.GL_RENDERBUFFER, GLES20.GL_DEPTH_COMPONENT16,
				this.renderTexture.getWidth(), this.renderTexture.getHeight());
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, 0);
	}

	public void freeResources() {
		final int[] rendBuff = new int[] { this.fbHandle };
		if (this.fbHandle > 0) {
			GLES20.glDeleteFramebuffers(1, rendBuff, 0);
			this.fbHandle = -1;
		}
		if (this.depthFbHandle > 0) {
			rendBuff[0] = this.depthFbHandle;
			GLES20.glDeleteRenderbuffers(1, rendBuff, 0);
			this.depthFbHandle = 1;
		}
	}

	public RenderTexture getSharedRenderTexture() {
		return this.renderTexture;
	}

	@Override
	public void onSurfaceChanged(int width, int height) {
		this.viewSize[0] = width;
		this.viewSize[1] = height;

		//updateRenderTextureDim(width, height);
		updateProj();
	}

	public final int[] getViewSize(int[] viewSize) {
		System.arraycopy(this.viewSize, 0, viewSize, 0, 2);
		return viewSize;
	}

	public final int getWidth() {
		return this.renderTexture.getWidth();
	}

	public final int getHeight() {
		return this.renderTexture.getHeight();
	}

	public final float[] getProjMatrix(float[] projMatrix) {
		System.arraycopy(this.projMatrix, 0, projMatrix, 0, 16);
		return projMatrix;
	}

	protected void updateProj() {
		this.renderingPrms.setViewSize(this.viewSize);

		float ratio =  ((float) this.viewSize[0]) / this.viewSize[1];
		perspectiveM(this.projMatrix, 0, this.fovY, ratio, this.nearClipPlane, this.farClipPlane);

		// Sets the projection matrix.
		this.renderingPrms.setProjMatrix(this.projMatrix);
	}

	public final void setBackgroundColor(float r, float g, float b, float a) {
		this.backgroundColor[0] = r;
		this.backgroundColor[1] = g;
		this.backgroundColor[2] = b;
		this.backgroundColor[3] = a;
	}

	protected void initGLParams() {
//		GLES20.glEnable(GLES20.GL_DEPTH_TEST);
//		GLES20.glClearDepthf(1.0f);
//		GLES20.glDepthFunc(GLES20.GL_LEQUAL);
//		GLES20.glDepthMask(true);

		// cull backface
		GLES20.glEnable(GLES20.GL_CULL_FACE);
		GLES20.glCullFace(GLES20.GL_BACK);

		this.glParamsInitialized = true;
	}

	/**
	 * When beginning a new cycle of rendering, this method let mark the current
	 * {@link RenderTexture} as not swapped (i.e. marked as not containing a
	 * previous rendering.
	 */
	public void emptyRenderTexture() {
		this.renderTexture.markAsNotSwapped();
	}

	public void setToFirstPass() {
		this.passNum = 0;
		this.needsPass = true;
	}

	public int getCurrentPass() {
		return this.passNum;
	}

	public boolean renderOnePassToTexture() {
		if (!this.needsPass) {
			return false;
		}

		preRenderToTexture();

		// will be updated once a rendering occurred.
		this.needsPass = false;

		// Sets the view matrix (i.e. inverse of the camera transform).
		getSharedTransform().getInverseWorldMatrix4(this.viewMatrix);
		this.renderingPrms.setViewMatrix(this.viewMatrix);

		// Proj x view
		Matrix.multiplyMM(this.viewProjMatrix, 0, this.projMatrix, 0, this.viewMatrix, 0);
		this.renderingPrms.setViewProjMatrix(this.viewProjMatrix);

		final Camera currentCamera = this;

		if (this.passNum == 0) {
			// First pass (i.e. pass 0).
			SceneConductor.getInstance().visitEachSceneObject(new SceneConductor.ISceneObjectVisitor() {
				@Override
				public void visitSceneObject(SceneObject sob) {
					if (sob.isEnabled()) {
						final Renderer renderer = sob.getSharedRenderer();
						if (renderer != null) {
							renderer.setToFirstPass();
							renderer.renderOnePass(currentCamera);
							if (renderer.stillNeedsPass()) {
								multiPassSOb.add(sob);
							}
						}
					}
				}
			});
			this.needsPass = (this.multiPassSOb.size() > 0);
		} else {
			final int nbSObs = this.multiPassSOb.size();
			for (int iSOb = 0; iSOb < nbSObs; iSOb++) {
				final SceneObject sob = this.multiPassSOb.get(iSOb);
				if (sob != null) {
					final Renderer renderer = sob.getSharedRenderer();
					renderer.renderOnePass(currentCamera);
					if (!renderer.stillNeedsPass()) {
						// This sob does not need other passes, so discard it.
						this.multiPassSOb.set(iSOb, null);
					} else {
						this.needsPass = true;
					}
				}
			}
		}

		this.renderTexture.swapPrevAndCurrentTexture();

		// Prepare for the next run.
		this.passNum++;

		if (!needsPass) {
			this.multiPassSOb.clear();
		}

		return this.needsPass;
	}

	public boolean stillNeedsPass() {
		return this.needsPass;
	}

	private void preRenderToTexture() {
		Texture.resetTextureCounter();

		if (this.fbHandle < 1) {
			throw new WrongFrameBufferHandleException(
					"FBO not initialized? (fbHandle = " + this.fbHandle + ")");
		}

		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, this.fbHandle);

		GLES20.glFramebufferTexture2D(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_COLOR_ATTACHMENT0,
                GLES20.GL_TEXTURE_2D, this.renderTexture.getTextureHandle(), 0);

		GLES20.glFramebufferRenderbuffer(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_DEPTH_ATTACHMENT,
				GLES20.GL_RENDERBUFFER, this.depthFbHandle);

		final int fbStatus = GLES20.glCheckFramebufferStatus(GLES20.GL_FRAMEBUFFER);
		if (fbStatus != GLES20.GL_FRAMEBUFFER_COMPLETE) {
			throw new FrameBufferIncompleteException(
					"Could not bind frame buffer: " + fbStatus);
		}

		if (!this.glParamsInitialized) {
			initGLParams();
		}

    	GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);

		// Set the background frame color
		GLES20.glClearColor(
				this.backgroundColor[0],
				this.backgroundColor[1],
				this.backgroundColor[2],
				this.backgroundColor[3]);

		GLES20.glViewport(0, 0, this.renderTexture.getWidth(), this.renderTexture.getHeight());
	}

	/**
	 * Gets parameters that can be passed to all shaders for a rendering with
	 * this {@link Camera}. Note that info contained by the returned object will
	 * be updated <b>only</b> by the method <code>render</code>.
	 *
	 * @return an object that contains info for rendering.
	 */
	public final CommonRenderingPrms getSharedCommonRenderingPrms() {
		return this.renderingPrms;
	}

	/**
     * Define a projection matrix in terms of a field of view angle, an
     * aspect ratio, and z clip planes.
     * This code has been extracted from Android API level 14.
     *
     * @param m the float array that holds the perspective matrix
     * @param offset the offset into float array m where the perspective
     * matrix data is written
     * @param fovy field of view in y direction, in degrees
     * @param aspect width to height aspect ratio of the viewport
     * @param zNear
     * @param zFar
     */
    public static void perspectiveM(float[] m, int offset,
          float fovy, float aspect, float zNear, float zFar) {
        float f = 1.0f / (float) Math.tan(fovy * (Math.PI / 360.0));
        float rangeReciprocal = 1.0f / (zNear - zFar);

        m[offset + 0] = f / aspect;
        m[offset + 1] = 0.0f;
        m[offset + 2] = 0.0f;
        m[offset + 3] = 0.0f;

        m[offset + 4] = 0.0f;
        m[offset + 5] = f;
        m[offset + 6] = 0.0f;
        m[offset + 7] = 0.0f;

        m[offset + 8] = 0.0f;
        m[offset + 9] = 0.0f;
        m[offset + 10] = (zFar + zNear) * rangeReciprocal;
        m[offset + 11] = -1.0f;

        m[offset + 12] = 0.0f;
        m[offset + 13] = 0.0f;
        m[offset + 14] = 2.0f * zFar * zNear * rangeReciprocal;
        m[offset + 15] = 0.0f;
    }
}
