package com.laguerre.kan;

import java.util.ArrayList;

import android.opengl.GLES20;
import android.util.SparseArray;

public class MeshRenderer extends Renderer {
	private final SparseArray<ArrayList<Material>> materialsPerSubMesh =
			new SparseArray<ArrayList<Material>>();

	protected int passNum = 0;
	protected boolean needsPass = true;

	public final void addMaterial(Material mat) {
		addMaterial(mat, 0);
	}

	public final void removeMaterial(Material mat) {
		removeMaterial(mat, 0);
	}

	public final void addMaterial(Material mat, int subMeshId) {
		ArrayList<Material> materials = this.materialsPerSubMesh.get(subMeshId);
		if (materials == null) {
			materials = new ArrayList<Material>();
			this.materialsPerSubMesh.put(subMeshId, materials);
		}
		if (!materials.contains(mat)) {
			materials.add(mat);
		}
	}

	public final void removeMaterial(Material mat, int subMeshId) {
		ArrayList<Material> materials = this.materialsPerSubMesh.get(subMeshId);
		if (materials != null) {
			materials.remove(mat);
		}
	}

	// Used only by render() to avoid many memory allocations.
	protected final float[] renderBufferMat4 = new float[16];
	protected final float[] renderBufferInvMat4 = new float[16];

	@Override
	public void setToFirstPass() {
		this.passNum = 0;
		this.needsPass = true;
	}

	@Override
	public int getCurrentPass() {
		return this.passNum;
	}

	@Override
	public boolean renderOnePass(Camera camera) {
		if (!this.needsPass) {
			return false;
		}

		// will be updated once a rendering occurred.
		this.needsPass = false;

		final SceneObject parentSOb = getSharedSOb();
		final Mesh mesh = parentSOb.getAnyComponent(Mesh.class);
		if (mesh != null) {
			final Transform t = parentSOb.getSharedTransform();

			// Shader values that are common to every shader of this SOb.
			final Shader.ModelRelativeData shaderValues = new Shader.ModelRelativeData();

			// Model matrix and the inverse
			t.getWorldMatrix4(this.renderBufferMat4);
			t.getInverseWorldMatrix4(this.renderBufferInvMat4);
			shaderValues.setModelMatrix(this.renderBufferMat4, this.renderBufferInvMat4);

			// Other shader values that are camera dependent only.
			final CommonRenderingPrms renderingPrms = camera.getSharedCommonRenderingPrms();

			// View-model matrix and the inverse
			Transform.combine(this.renderBufferMat4, renderingPrms.getSharedViewMatrix(), shaderValues.getSharedModelMatrix());
			Transform.combine(this.renderBufferInvMat4, shaderValues.getSharedInvModelMatrix(), renderingPrms.getSharedInvViewMatrix());
			shaderValues.setModelViewMatrix(this.renderBufferMat4, this.renderBufferInvMat4);

			// Proj-view-model matrix and the inverse
			Transform.combine(this.renderBufferMat4, renderingPrms.getSharedProjMatrix(), shaderValues.getSharedModelViewMatrix());
			Transform.combine(this.renderBufferInvMat4, shaderValues.getSharedInvModelViewMatrix(), renderingPrms.getSharedInvProjMatrix());
			shaderValues.setModelViewProjModelMatrix(this.renderBufferMat4, this.renderBufferInvMat4);

			final Vector3 eyePosition = camera.getSharedTransform().getWorldTranslation();

			final int nbOfSubMeshes = mesh.getNbOfSubMeshes();
			for (int i = 0; i < nbOfSubMeshes; i++) {
				final ArrayList<Material> materials = this.materialsPerSubMesh.get(i);
				if (materials != null) {
					// Initialize the texture counter for all the Materials of this sub-mesh.
					Texture.resetTextureCounter();

					for (Material mat : materials) {
						if (mat.isPassAvailable(this.passNum)) {
							mat.setShaderVector("viewSize", renderingPrms.getSharedViewSize());
							//
							mat.setShaderMatrix4("V", renderingPrms.getSharedViewMatrix());
							mat.setShaderMatrix4("invV", renderingPrms.getSharedInvViewMatrix());
							//
							mat.setShaderMatrix4("P", renderingPrms.getSharedProjMatrix());
							mat.setShaderMatrix4("invP", renderingPrms.getSharedInvProjMatrix());
							//
							mat.setShaderMatrix4("VP", renderingPrms.getSharedViewProjMatrix());
							mat.setShaderMatrix4("invVP", renderingPrms.getSharedInvViewProjMatrix());
							//
							mat.setShaderMatrix4("M", shaderValues.getSharedModelMatrix());
							mat.setShaderMatrix4("invM", shaderValues.getSharedInvModelMatrix());
							//
							mat.setShaderMatrix4("MV", shaderValues.getSharedModelViewMatrix());
							mat.setShaderMatrix4("invMV", shaderValues.getSharedInvModelViewMatrix());
							//
							mat.setShaderMatrix4("MVP", shaderValues.getSharedModelViewProjMatrix());
							mat.setShaderMatrix4("invMVP", shaderValues.getSharedInvModelViewProjMatri());
							//
							mat.setShaderVector("eyePosition", eyePosition);
							//
							mat.setShaderBuffer("position",
									4,
									mesh.getVertexStride(i),
									mesh.getSharedVertexBuffer(i));
							//
							mat.setShaderBuffer("normal",
									3,
									mesh.getNormalStride(i),
									mesh.getSharedNormalBuffer(i));
							//
							mat.setShaderBuffer("uv",
									2,
									mesh.getUVStride(i),
									mesh.getSharedUVBuffer(i));
							//
							mat.setShaderValue(
									"hasRenderedTex",
									camera.getSharedRenderTexture().hasSwapped());
							mat.setShaderTexture(
									"prevRenderedTex",
									camera.getSharedRenderTexture().getPrevTextureHandle());

							mat.setShaderVector("ambientLight",
									RenderingSettings.getInstance().getSharedAmbientLight());

							Lights.getInstance().passLightsToShader(mat.getSharedShader());

							mat.setPass(this.passNum);

							mat.use();

							// Still passes required?
							final int nbPasses = mat.passCount();
							// 1st pass = 0!
							if (this.passNum + 1 < nbPasses) {
								this.needsPass = true;
							}
						}
					}
				}

		    	GLES20.glDrawElements(
		    			GLES20.GL_TRIANGLES, mesh.getDrawListCount(i),
		                GLES20.GL_UNSIGNED_SHORT, mesh.getSharedDrawListBuffer(i));

				if (materials != null) {
					for (Material mat : materials) {
						mat.postRender(mesh);

						mat.releaseAllShaderBuffers();
					}
				}
			}
		}

		// Prepare for the next run.
		this.passNum++;

		return this.needsPass;
	}

	@Override
	public boolean stillNeedsPass() {
		return this.needsPass;
	}
}
